class System:
    initial_state = None

    def calculate_step(self, state, inp):
        return (state, inp)


class SystemSimulator:
    def __init__(self, system):
        self.system = system
        self.reset()

    def step(self, inp):
        new_state, out = self.system.calculate_step(self.state, inp)
        self.state = new_state
        return out

    def reset(self):
        self.state = self.system.initial_state

    def get_response(self, inputs, reset=True):
        if reset:
            self.reset()
        return [self.step(inp) for inp in inputs]


class R(System):
    def __init__(self, output0=0):
        self.initial_state = None # modify if necessary

    def calculate_step(self, state, inp):
        # your code here
        return (new_state, output)


class Gain(System):
    def __init__(self, k):
        self.initial_state = None # modify if necessary

    def calculate_step(self, state, inp):
        # your code here
        return (new_state, output)


class FeedforwardAdd(System):
    def __init__(self, s1, s2):
        self.initial_state = None # modify if necessary

    def calculate_step(self, state, inp):
        # your code here
        return (new_state, output)


class Cascade(System):
    def __init__(self, s1, s2):
        self.initial_state = None # modify if necessary

    def calculate_step(self, state, inp):
        # your code here
        return (new_state, output)


class FeedbackAdd(System):
    def __init__(self, s1, s2):
        self.initial_state = (s1.initial_state, s2.initial_state)
        self.s1 = s1
        self.s2 = s2

    def calculate_step(self, state, inp):
        s1_state, s2_state = state

        # Here we encounter a problem.  It seems like s1's output
        # depends on s2's output, and s2's output depends on s1's
        # output.  Which one do we run first?  It looks like getting
        # either output is impossible!
        #
        # Importantly, we will make the assumption that at least one of s1
        # and s2 has the property that its output at time n does not depend
        # on its input at time n (depends only on older inputs/outputs).
        # The net effect of this is that s2's output at time n does not
        # depend on the input to the system at time n.
        #
        # We need to know s2's output in order to compute s1's output (and
        # to know how to update s2's state).  We will get around this by
        # using a "fake" input to determine s2's output (which we know was
        # unaffected by our bogus input), and then using that result to
        # compute the real outputs and states of the two internal systems.

        # Firstly, we propagate a bogus input through the system in order
        # to find s2's output.
        s1_state_hyp, out1_hyp = self.s1.calculate_step(s1_state, 99999999)
        s2_state_hyp, out2_hyp = self.s2.calculate_step(s2_state, out1_hyp)

        # Because of the property mentioned above, we know that s2's output
        # did not depend on the bogus input, so this is the right value.
        out2_real = out2_hyp

        # Knowing s2's output, we can now know the actual value that we should
        # pass in to s1.  Now we can compute the outputs (and new states) of
        # s1 and s2 correctly.
        s1_real_inp = inp + out2_real
        s1_state_real, out1_real = self.s1.calculate_step(s1_state, s1_real_inp)
        s2_state_real, out2_real = self.s2.calculate_step(s2_state, out1_real)

        new_state = (s1_state_real, s2_state_real)
        output = out1_real

        return (new_state, output)